import json
from typing import Dict, Union, List

from django.forms import ClearableFileInput
from django.http import QueryDict
from django.template.loader import render_to_string
from django.utils.datastructures import MultiValueDict
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django_file_form.util import get_list

from django_file_form.models import (
    PlaceholderUploadedFile,
    S3UploadedFileWithId,
    UploadedFileWithId,
)

TRANSLATIONS = {
    "Cancel": _("Cancel"),
    "Delete": _("Delete"),
    "Delete failed": _("Delete failed"),
    "Upload failed": _("Upload failed"),
    "Drop your files here": _("Drop your files here"),
}

UploadedFileTypes = Union[
    PlaceholderUploadedFile,
    UploadedFileWithId,
    List[Union[PlaceholderUploadedFile, UploadedFileWithId]],
]


def get_uploaded_files(value: UploadedFileTypes):
    if not value:
        return []

    return [
        file_info.get_values()
        if hasattr(file_info, "file_id")
        else dict(name=file_info.name)
        for file_info in get_list(value)
        if not getattr(file_info, "is_placeholder", False)
        and not getattr(file_info, "is_s3direct", False)
    ]


def get_placeholder_files(data: QueryDict, field_name: str):
    placeholder_field_name = field_name + "-placeholder"
    value = data.get(placeholder_field_name)

    if not value:
        return []
    else:
        return [
            PlaceholderUploadedFile(
                name=placeholder["name"],
                size=placeholder["size"],
                file_id=placeholder["id"],
            )
            for placeholder in json.loads(value)
        ]


def get_s3_uploaded_files(data: QueryDict, field_name: str):
    s3uploaded_field_name = field_name + "-s3direct"
    value = data.get(s3uploaded_field_name)

    if not value:
        return []
    else:
        return [
            S3UploadedFileWithId(
                name=s3uploaded["name"],
                original_name=s3uploaded["original_name"],
                size=s3uploaded["size"],
                file_id=s3uploaded["id"],
            )
            for s3uploaded in json.loads(value)
        ]


def get_file_meta(data: QueryDict, field_name: str):
    meta_field_name = field_name + "-metadata"
    value = data.get(meta_field_name)
    if not value:
        return {}
    try:
        res = json.loads(value)
        if not isinstance(res, dict):
            return {}
        return res
    except Exception:
        return {}


class BaseUploadWidget(ClearableFileInput):
    def render(
        self,
        name: str,
        value: UploadedFileTypes,
        attrs=None,
        renderer=None,
    ):
        upload_input = super().render(name, value, attrs, renderer)
        return mark_safe(
            render_to_string(
                "django_file_form/upload_widget.html",
                dict(
                    input=upload_input,
                    translations=json.dumps(TRANSLATIONS),
                    uploaded_files=json.dumps(get_uploaded_files(value)),
                ),
            )
        )


class UploadWidget(BaseUploadWidget):
    def value_from_datadict(self, data: QueryDict, files: MultiValueDict, name: str):
        upload = super().value_from_datadict(data, files, name)

        if upload:
            return upload
        else:
            placeholders = get_placeholder_files(data, name)
            s3uploaded = get_s3_uploaded_files(data, name)
            metadata = get_file_meta(data, name)
            obj = (
                placeholders[0]
                if placeholders
                else (s3uploaded[0] if s3uploaded else None)
            )
            if obj is not None and obj.name in metadata:
                obj.metadata = metadata[obj.name]
            return obj


class UploadMultipleWidget(BaseUploadWidget):
    def value_from_datadict(
        self, data: QueryDict, files: Union[Dict, MultiValueDict], name: str
    ):
        if hasattr(files, "getlist"):
            metadata = get_file_meta(data, name)
            objs = (
                files.getlist(name)
                + get_placeholder_files(data, name)
                + get_s3_uploaded_files(data, name)
            )
            for obj in objs:
                if obj.name in metadata:
                    obj.metadata = metadata[obj.name]
            return objs
        else:
            # NB: django-formtools wizard uses dict instead of MultiValueDict
            return super().value_from_datadict(data, files, name)



import os
import uuid
from pathlib import Path

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files import uploadedfile
from django.core.files.storage import FileSystemStorage
from django.db import models
from django.utils import timezone
from django.utils.module_loading import import_string

from .util import ModelManager, get_upload_path


class TemporaryUploadedFileManager(ModelManager):
    def delete_unused_files(self, delete=True, now=None):
        deleted_files = []

        for t in self.get_queryset():
            if t.must_be_deleted(now):
                deleted_files.append(Path(t.uploaded_file.name).name)

                if delete:
                    t.delete()

        temp_path = get_upload_path()

        for f in temp_path.iterdir():
            basename = f.name

            if not self.get_for_file(basename):
                deleted_files.append(basename)

                if delete:
                    f.unlink()

        return deleted_files

    def get_for_file(self, filename):
        return self.try_get(uploaded_file=get_upload_to_for_filename(filename))


def get_upload_to_for_filename(filename):
    path = getattr(settings, "FILE_FORM_UPLOAD_DIR", "temp_uploads")
    return os.path.join(path, filename)


def get_upload_to(_instance, filename):
    return get_upload_to_for_filename(filename)


if hasattr(settings, "FILE_FORM_TEMP_STORAGE"):
    storage_class = import_string(settings.FILE_FORM_TEMP_STORAGE)
else:
    storage_class = FileSystemStorage


class TemporaryUploadedFile(models.Model):
    """
    TemporaryUploadedFile is used for temporary storage of an uploaded file.

    * Added when a file is uploaded using the tus ajax upload
    * Removed when the form is submitted sucessfully
    * Or removed later by 'python manage.py delete_unused_files'
    """
    created = models.DateTimeField(default=timezone.now)
    uploaded_file = models.FileField(
        max_length=255, storage=storage_class(), upload_to=get_upload_to
    )
    original_filename = models.CharField(max_length=255)
    field_name = models.CharField(max_length=255, null=True, blank=True)
    file_id = models.CharField(max_length=40)
    form_id = models.CharField(max_length=40)

    objects = TemporaryUploadedFileManager()

    class Meta(object):
        # Query string to get back existing uploaded file is using form_id and field_name
        index_together = (("form_id", "field_name"),)
        db_table = "django_file_form_uploadedfile"

    def __str__(self):
        return str(self.original_filename or "")

    def delete(self, *args, **kwargs):
        if self.uploaded_file and self.uploaded_file.storage.exists(
            self.uploaded_file.name
        ):
            self.uploaded_file.delete()

        super().delete(*args, **kwargs)

    def must_be_deleted(self, now=None):
        now = now or timezone.now()

        return (now - self.created).days >= 1

    def get_uploaded_file(self):
        return UploadedFileWithId(
            file=self.uploaded_file, name=self.original_filename, file_id=self.file_id
        )


class UploadedFileWithId(uploadedfile.UploadedFile):
    def __init__(self, file_id, metadata=None, **kwargs):
        super().__init__(**kwargs)

        self.file_id = file_id
        self.size = self.file.size

        self.is_placeholder = False
        self.is_s3direct = False
        self.metadata = metadata

    def get_values(self):
        return dict(
            id=self.file_id, name=self.name, size=self.size, metadata=self.metadata
        )


class PlaceholderUploadedFile(object):
    def __init__(self, name, file_id=None, size=None, metadata=None):
        self.name = name
        self.file_id = file_id or uuid.uuid4().hex
        if size is None:
            self.size = os.path.getsize(self.name)
        else:
            self.size = size

        self.is_placeholder = True
        self.is_s3direct = False
        self.metadata = metadata

    def get_values(self):
        return dict(
            id=self.file_id,
            placeholder=True,
            name=self.name,
            size=self.size,
            metadata=self.metadata,
        )


try:
    from storages.backends.s3boto3 import S3Boto3Storage, S3Boto3StorageFile
    from storages.utils import setting

    class S3UploadedFileWithId(S3Boto3StorageFile):
        def __init__(self, file_id, name, original_name, size, metadata=None, **kwargs):
            boto_storage = S3Boto3Storage(
                bucket_name=setting("AWS_STORAGE_BUCKET_NAME"),
                endpoint_url=setting("AWS_S3_ENDPOINT_URL"),
                access_key=setting(
                    "AWS_S3_ACCESS_KEY_ID", setting("AWS_ACCESS_KEY_ID")
                ),
                secret_key=setting(
                    "AWS_S3_SECRET_ACCESS_KEY", setting("AWS_SECRET_ACCESS_KEY")
                ),
            )

            super().__init__(name=name, mode="rb", storage=boto_storage, **kwargs)

            self.file_id = file_id
            self.original_name = original_name
            # self.size is derived from S3boto3StorageFile
            # but size is passed for consistency, and potentially
            # for validation
            self.is_placeholder = False
            self.is_s3direct = True
            self.metadata = metadata

        def get_values(self):
            return dict(
                id=self.file_id,
                placeholder=False,
                name=self.name,
                size=self.size,
                metadata=self.metadata,
            )


except (ImportError, ImproperlyConfigured):
    # S3 is an optional feature but we keep the symbol
    class S3UploadedFileWithId(object):
        pass



import json
import uuid

from django.urls import reverse
from django.forms import CharField, Form, HiddenInput

# UploadedFileField and MultipleUploadedFileField must be in this module because they are in the api
from .fields import UploadedFileField, MultipleUploadedFileField
from .util import get_list


class FileFormMixin(object):
    def __init__(self, *args, **kwargs):
        s3_upload_dir = kwargs.pop(
            "s3_upload_dir",
            self.s3_upload_dir if hasattr(self, "s3_upload_dir") else None,
        )
        super().__init__(*args, **kwargs)

        if s3_upload_dir is not None:
            self.add_hidden_field("s3_upload_dir", s3_upload_dir)
            self.add_s3_uploaded_files()
            self.add_hidden_field("upload_url", reverse("s3_upload"))
        else:
            self.add_hidden_field("upload_url", reverse("tus_upload"))

        self.add_hidden_field("form_id", uuid.uuid4())
        self.add_placeholder_inputs()
        self.add_metadata_inputs()

    def add_hidden_field(self: Form, name: str, initial):
        self.fields[name] = CharField(
            widget=HiddenInput, initial=initial, required=False
        )

    def full_clean(self: Form):
        if not self.is_bound:
            # Form is unbound; just call super
            super().full_clean()
        else:
            # Update file data of the form
            self.update_files_data()

            # Call super
            super().full_clean()

    def file_form_field_names(self: Form):
        return [
            field_name
            for field_name, field in self.fields.items()
            if hasattr(field, "get_file_data")
        ]

    def update_files_data(self: Form):
        form_id = self.data.get(self.add_prefix("form_id"))

        if form_id:
            for field_name in self.file_form_field_names():
                field = self.fields[field_name]
                prefixed_field_name = self.add_prefix(field_name)
                file_data = field.get_file_data(prefixed_field_name, form_id)

                if file_data:
                    # NB: django-formtools wizard uses dict instead of MultiValueDict
                    if isinstance(file_data, list) and hasattr(self.files, "setlist"):
                        self.files.setlist(prefixed_field_name, file_data)
                    else:
                        self.files[prefixed_field_name] = file_data

    def delete_temporary_files(self: Form):
        form_id = self.data.get(self.add_prefix("form_id"))

        if form_id:
            for field_name, field in self.fields.items():
                if hasattr(field, "delete_file_data"):
                    prefixed_field_name = self.add_prefix(field_name)
                    field.delete_file_data(prefixed_field_name, form_id)

    def add_s3_uploaded_files(self):
        for field_name in self.file_form_field_names():
            s3_uploaded_field_name = f"{field_name}-s3direct"
            # the field is always empty initially
            self.add_hidden_field(s3_uploaded_field_name, "[]")

    def add_placeholder_inputs(self):
        for field_name in self.file_form_field_names():
            placeholder_field_name = f"{field_name}-placeholder"
            self.add_hidden_field(
                placeholder_field_name,
                json.dumps(self.get_placesholders_for_field(field_name)),
            )

    def get_placesholders_for_field(self: Form, field_name: str):
        initial_values = get_list(self.initial.get(field_name, []))

        return [
            value.get_values()
            for value in initial_values
            if getattr(value, "is_placeholder")
        ]

    def add_metadata_inputs(self):
        for field_name in self.file_form_field_names():
            metadata_field_name = f"{field_name}-metadata"
            self.add_hidden_field(
                metadata_field_name, json.dumps(self.get_metadata_for_field(field_name))
            )

    def get_metadata_for_field(self: Form, field_name: str):
        initial_values = get_list(self.initial.get(field_name, []))
        return {
            value.name: value.metadata
            for value in initial_values
            if value.metadata is not None
        }



from django.contrib import admin

from . import models


class TemporaryUploadedFileAdmin(admin.ModelAdmin):
    list_display = ["original_filename", "created"]
    date_hierarchy = "created"


admin.site.register(models.TemporaryUploadedFile, TemporaryUploadedFileAdmin)
